@@ -1,5 +1,5 @@
 /*
- * Copyright 2006-2007 TSUBAKIMOTO Hiroya <zorac@4000do.co.jp>
+ * Copyright 2006-2008 TSUBAKIMOTO Hiroya <zorac@4000do.co.jp>
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License
@@ -128,11 +128,11 @@ class console
 {
   static struct state_t
   {
+    DWORD id;
     HWND hwnd;
-    HANDLE done;
-    bool shutdown;
+    HANDLE close;
 
-    state_t() : hwnd(NULL), done(NULL), shutdown(false) {}
+    state_t() : id(GetCurrentThreadId()), hwnd(NULL), close(NULL) {}
   } _state;
 
   static BOOL CALLBACK _handler(DWORD type)
@@ -140,10 +140,13 @@ class console
     switch (type) {
     case CTRL_LOGOFF_EVENT:
     case CTRL_SHUTDOWN_EVENT:
-      _state.shutdown = true;
+      _state.hwnd = NULL;
       break;
     }
-    return _state.done && SetEvent(_state.done);
+    if (PostThreadMessage(_state.id, WM_QUIT, 0, 0)) {
+      WaitForSingleObject(_state.close, 20 * 1000UL);
+    }
+    return FALSE;
   }
 
   static BOOL CALLBACK _find(HWND hwnd, LPARAM lParam)
@@ -170,7 +173,7 @@ class console
 public:
   static void open()
   {
-    _state.done = complete(CreateEvent(NULL, TRUE, FALSE, NULL));
+    _state.close = complete(CreateEvent(NULL, TRUE, FALSE, NULL));
     complete(AllocConsole() && SetConsoleCtrlHandler(_handler, TRUE));
     EnumWindows(_find, LPARAM(GetCurrentProcessId()));
     _stdio(stdin, 0, GetStdHandle(STD_INPUT_HANDLE));
@@ -180,8 +183,9 @@ public:
 
   static void close()
   {
+    SetEvent(_state.close);
     GenerateConsoleCtrlEvent(CTRL_C_EVENT, 0);
-    PostMessage(_state.hwnd, WM_CLOSE, 0, 0);
+    FreeConsole();
   }
 
   static void show(bool show = true)
@@ -199,16 +203,19 @@ public:
 
   static bool idle()
   {
-    while (MsgWaitForMultipleObjects
-	   (1, &_state.done, FALSE, INFINITE, QS_ALLINPUT) != WAIT_OBJECT_0) {
+    for (;;) {
       MSG msg;
-      while (PeekMessage(&msg, NULL, 0, 0, PM_REMOVE)) {
-	if (msg.message == WM_QUIT) return true;
+      switch (GetMessage(&msg, NULL, 0, 0)) {
+      default:
 	TranslateMessage(&msg);
 	DispatchMessage(&msg);
+	continue;
+      case 0: break;
+      case -1: throw GetLastError();
       }
+      break;
     }
-    return !_state.shutdown;
+    return _state.hwnd;
   }
 };
 console::state_t console::_state;
@@ -450,8 +457,16 @@ public:
       int i = 0;
       std::vector<string_t>::iterator p;
       for (p = menus.begin(); p != menus.end(); ++p) {
-	_cmds.push_back(config(MENU, *p));
-	complete(InsertMenu(_hmenu, i, MF_BYPOSITION, ID_MENU_USER + i, p->c_str()));
+	unsigned flag = MF_BYPOSITION;
+	string_t cmd;
+	if (p->size() == 1 && (*p)[0] == _T('-')) {
+	  flag |= MF_SEPARATOR;
+	} else {
+	  cmd = config(MENU, *p);
+	}
+	if (cmd.empty()) flag |= MF_GRAYED;
+	_cmds.push_back(cmd);
+	complete(InsertMenu(_hmenu, i, flag, ID_MENU_USER + i, p->c_str()));
 	++i;
       }
       complete(InsertMenu(_hmenu, i, MF_BYPOSITION | MF_SEPARATOR, 0, NULL));
@@ -612,26 +627,28 @@ int WINAPI WinMain(HINSTANCE, HINSTANCE, LPSTR, int)
   if (!CreateMutex(NULL, TRUE, mutex) ||
       GetLastError() == ERROR_ALREADY_EXISTS) return 0;
   try {
-    static const TCHAR OPTION[] = TEXT("option");
-    static const TCHAR CRON[] = TEXT("cron");
     console::open();
-    config_t config(TEXT("sws.ini"));
-    drive_t drive(config(OPTION, TEXT("drive")),
-		  config.path(OPTION, TEXT("basedir")));
-    menu_t menu(MAKEINTRESOURCE(1));
-    trayicon_t icon(MAKEINTRESOURCE(1), menu.load(config));
-    task_t::config(config);
-    task_t::environment(TEXT("env"));
-    task_t::timer(CRON, config(OPTION, CRON, 10) * 60000);
-    _tprintf(TEXT("Boot:\n"));
-    task_t(TEXT("boot")).execute();
-    console::show(false);
-    if (console::idle()) {
-      console::show();
-      _tprintf(TEXT("Shutdown:\n"));
-      task_t(TEXT("shutdown")).execute();
-      console::close();
+    {
+      static const TCHAR OPTION[] = TEXT("option");
+      static const TCHAR CRON[] = TEXT("cron");
+      config_t config(TEXT("sws.ini"));
+      drive_t drive(config(OPTION, TEXT("drive")),
+		    config.path(OPTION, TEXT("basedir")));
+      menu_t menu(MAKEINTRESOURCE(1));
+      trayicon_t icon(MAKEINTRESOURCE(1), menu.load(config));
+      task_t::config(config);
+      task_t::environment(TEXT("env"));
+      task_t::timer(CRON, config(OPTION, CRON, 10) * 60000);
+      _tprintf(TEXT("Boot:\n"));
+      task_t(TEXT("boot")).execute();
+      console::show(false);
+      if (console::idle()) {
+	console::show();
+	_tprintf(TEXT("Shutdown:\n"));
+	task_t(TEXT("shutdown")).execute();
+      }
     }
+    console::close();
     return 0;
   } catch (DWORD err) {
     _ftprintf(stderr, TEXT("Error=%lu\n"), err);
